package RPM::Source::Convert::Plugins::DistroMap::Dependencies;
use strict;
use warnings;

use RPM::Source::Tools::Dependency::Filter;
use RPM::Source::Transformation::Factory::DependencyFilter;

&RPM::Source::Transformation::Factory::DependencyFilter::register_filter(
    -name => 'DistroMap Binary Flags',
    -description => 'filter that implements DistroMap binary flags',
    -class => 'RPM::Source::Tools::Dependency::Filter::DistroMapBinaryFlags',
    -priority => FILTER_PRIORITY_DISTROMAP,
);
&RPM::Source::Transformation::Factory::DependencyFilter::register_filter(
    -name => 'DistroMap Binary Names',
    -description => 'filter that implements DistroMap binary package name substitutions',
    -class => 'RPM::Source::Tools::Dependency::Filter::DistroMap',
    -priority => FILTER_PRIORITY_DISTROMAP,
);
&RPM::Source::Transformation::Factory::DependencyFilter::register_filter(
    -name => 'Drop Undefined Version Macro',
    -description => 'drop version if it is written as macro that is not defined in spec',
    -class => 'RPM::Source::Tools::Dependency::Filter::DropUndefinedVersionMacro',
    -priority => FILTER_PRIORITY_POSTFILTER,
);


package RPM::Source::Tools::Dependency::Filter::DropUndefinedVersionMacro;

use strict;
use warnings;
our @ISA=q/RPM::Source::Tools::Dependency::Filter/;
our $verbose=1;

sub apply {
    my ($self, $tagobj, $dep) = @_;
    my $version=$dep->{-version};
    return $dep unless defined $version;
    if ($version=~/^\%/) {
	my $expanded=$self->{-spec}->macros->macro_subst($version);
	if ($expanded eq '') {
	    print STDERR "info: dropped undefined $dep->{-version} in $dep->{-name} $dep->{-relation} $dep->{-version}\n" if $verbose;
	    delete $dep->{-name};
	    return;
	}
    }
    return $dep;
}

1;

package RPM::Source::Tools::Dependency::Filter::DistroMapBinaryFlags;

use strict;
use warnings;
use RPM::Source::Shared::Resource::DistroMap;

our @ISA=q/RPM::Source::Tools::Dependency::Filter/;
our $verbose=0;

sub apply {
    my ($self, $tagobj, $dep) = @_;
    my $version=$dep->{-version};
    my $distromap=$self->{-resourceobj}->get_distromap;
    my $expanded_name=$self->{-spec}->macros()->macro_subst($dep->{-name});
    if ($distromap->flag('flags/binary/exclude',$expanded_name)) {
	print STDERR "info: flags/binary/exclude: dropped $expanded_name dependency\n" if $verbose;
	delete $dep->{-name};
	return;
    }
    return $dep unless defined $version;
    if ($distromap->flag('flags/binary/noversion',$expanded_name)) {
	print STDERR "info: flags/binary/noversion: dropped $dep->{-version} in $expanded_name $dep->{-relation} $dep->{-version}\n" if $verbose;
	delete $dep->{-version};
	delete $dep->{-relation};
	return $dep;
    }
    if ($version=~/^\d+:/ and $distromap->flag('flags/binary/noepoch',$expanded_name)) {
	print STDERR "info: flags/binary/noepoch: dropped $dep->{-version} in $expanded_name $dep->{-relation} $dep->{-version}\n" if $verbose;
	$dep->{-version}=~s/^\d+://;
    }
    return $dep;
}

package RPM::Source::Tools::Dependency::Filter::DistroMap;
use strict;
use warnings;
our @ISA=q/RPM::Source::Tools::Dependency::Filter/;
our $verbose=1;

sub apply {
    my ($self, $tagobj, $dep) = @_;
    my $resourceobj=$self->{-resourceobj};
    return $dep unless $resourceobj;
    my $distromap=$self->{-resourceobj}->get_distromap;
    return $dep unless $distromap;
    my $spec=$self->{-spec};
    my $pkgname=$dep->{-name};
    # filename dependencies that are currenly not in apt index
    if ($pkgname=~m!^/usr/include/!) {
	my $header=$pkgname;
	$header=~s,^/usr/include/,,;
	my @mapping=$distromap->dest()->translate('headers',$header);
	$dep->{-name}=$mapping[0] if @mapping and (@mapping ==1);
    } else {
	# like, 
	# %define req_list dep0, dep1, ...
        # Requires: %{req_list}
	# TODO: handle %(eval ...) above apply()
	my @expanded_names=split(/\s+/,$spec->macros()->macro_subst_if_defined($pkgname));
	my @mappings;
	foreach my $expanded_name (@expanded_names) {
	    my @mapping=$distromap->multimap_binary_origin2dest($expanded_name);
	    if ($RPM::Source::Transform::verbose>1) {
		my $mapname=join(' ',@mapping);
		my $originname="$expanded_name [$pkgname]";
		$originname=$pkgname if $expanded_name eq $pkgname;
		print STDERR "distromap: $originname -> $mapname\n" if $mapname;
		print STDERR "distromap: no map for $originname\n" unless $mapname;
	    }
	    if (@mapping) {
		push @mappings, @mapping;
	    } else {
		push @mappings, $expanded_name;
	    }
	}
	warn "Oops!Dependencies.pm: undefined mappings[0] for $pkgname, expanded_names=".join(' ',@expanded_names) if $verbose and not defined $mappings[0] and @expanded_names>1;
	# regen deps objects
	if (@mappings and 
	    (($mappings[0] ne $expanded_names[0]) or 
	     # append; like boost-devel unpack to boost-devel + boost-*-devel
	     (scalar @mappings ne scalar @expanded_names))) {
	    if (@mappings==1) {
		$dep->{-name}=$mappings[0];
	    } elsif (@mappings >1) {
		return map {RPM::Source::Tools::Dependency->new(-name=>$_)} @mappings;
	    }
	}
    }
    return $dep;
}

1;
