package RPM::Source::Convert::Plugins::Generic2ALT;

use strict;
use warnings;
use Carp;
use File::Spec;
use File::Basename;
use File::Path qw(make_path remove_tree);
use File::Temp qw/tempfile tempdir/;

use RPM::Source::Tools::Dependency::Filter;
use RPM::Source::Shared::Resource::DistroMap;
use RPM::Source::Transformation::Factory;
use RPM::Source::Transformation::OptHandlerWrapper;
use RPM::Source::Transformation::Factory::Buildreq;
use RPM::Source::Transformation::Factory::ParentI18NMerge;
use RPM::Source::Transformation::Factory::DependencyFilter;

use RPM::Source::Convert::Plugins::DistroMap::Groups;
use RPM::Source::Convert::Plugins::DistroMap::Dependencies;
use RPM::Source::Convert::Plugins::DistroMap::Packages;
use RPM::Source::Convert::Plugins::DistroMap::SourceAnalyzer;
use RPM::Source::Convert::Plugins::DistroMap::SubpackageRename;
use RPM::Source::Convert::Plugins::Alternatives2ALT;
# hack ... should be proper extender or filter
$DistroMap::enable_generic_stem_mapper=1;

our @ISA=qw/RPM::Source::Transformation::Factory/;

push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our $verbose=0;

my $opt_import_info=1; # create import.info file
# due to rpm4.0
our $opt_force_ascii=1;

our @LONGOPT=(
    'import-info!' => \$opt_import_info,
    'force-ascii!' => \$opt_force_ascii,
    );

sub longopt_usage {
    print "GenericImport options:\n";
    print "	--no-import-info  no not create import.info file\n";
}

sub transformations {
    return 
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_PRECONVERT,
	    -name => 'generic_preconvert',
	    -opt_handler_sub=>\&opt_handler_generic_preconvert),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_CONVERT,
	    -name => 'generic_convert',
	    -opt_handler_sub=>\&opt_handler_generic_convert);
}

&init_plugin();

sub init_plugin {
    $DistroMap::default_originrepo='fedora';
    $DistroMap::default_originbranch=undef;
    $DistroMap::default_destrepo='altlinux';
    $DistroMap::default_destbranch=undef;
}

&RPM::Source::Transformation::Factory::DependencyFilter::register_filter(
    -name => 'post/preun explicit script dependencies',
    -description => 'drop cleaned up post/preun explicit script dependencies',
    -class => 'Generic2ALT::DependencyFilter::PostScriptsDepsCleanup',
    -tags => SPEC_TAG_REQUIRES,
    -priority => FILTER_PRIORITY_PREDISTROMAP,
);

sub opt_handler_generic_preconvert {
    my (undef,$spec,$parent,$opt,$resourceobj)=@_;
    foreach my $sec ($spec->get_sections()) {
	my $t = $sec->get_type();
	if ($t eq 'package') {
	    $sec->map_body(sub{
		print STDERR "WARNING: $1: replaced with $_" if s,^(Recommends|Suggests):,Requires:,;
			   });
	}
    }
}

sub opt_handler_generic_convert {
    my (undef,$spec,$parent,$opt,$resourceobj)=@_;
    my $cleanroot_pattern='^\s*(?:/bin/)?(?:rm|\%{__rm}|\%__rm)\s+-rf?\s+(?:\$RPM_BUILD_ROOT|\%\{?rpm_build_root\}?|\%\{?buildroot\}?)\s*$';
    my $mainsec=$spec->main_section;
    my $buildreq=$resourceobj->buildreq;
    &alt_generic_check_ahead($spec,$mainsec);
    &alt_generic_check_zip($buildreq,$spec);
    &alt_generic_check_jni($buildreq,$spec);
    &alt_generic_check_alt_macro($spec);
    my $fedora_like_macrodir=0;

    foreach my $sec ($spec->get_sections()) {
	my $t = $sec->get_type();
	print STDERR "Generic2ALT: processing \%$t ",$sec->get_raw_package,"\n" if $verbose>2;
	if ($t eq 'package') {

	    $sec->apply_body(sub {
		local $_=shift;
		if (/^%global\s+(?:rpm_macros_dir|macrosdir)|\s+.*_rpmconfigdir/) {
		    $fedora_like_macrodir=1;
		    return '';		# exclude
		}
		if (/^(?:Packager|Vendor|Distribution|Build[Rr]oot):/) {
		    return '';		# exclude
		}
		return $_;
			     });

	    # encoding mismatch. alt assume 8 bit
	    # TODO: detect utf8?
	    #$sec->exclude('^Summary\(de\):');
	    &alt_generic_cleanup_headers($resourceobj,$spec,$sec);
	} elsif ($t eq 'description') {
	    my $l=$sec->get_flag('-l');
	    #$sec->delete() if $l and $l=~/^(?:de|fr)$/;
	    &alt_generic_fix_desc($spec, $sec) unless $l;
	} elsif ($t eq 'files') {
	    &alt_generic_cleanup_files($spec, $sec);
	    $sec->map_body(\&__convert_fedora_like_macrodir) if $fedora_like_macrodir;
	} elsif ($t eq 'prep' || $t eq 'build' || $t eq 'install') {
	    $sec->exclude($cleanroot_pattern);
	    if ($t eq 'prep') {
		# for tricks like perldoc -t perlartistic > Artistic
		if ($sec->match_body(qr'(?:^|\W)perldoc\s')) {
		    $buildreq->add_buildreq('perldoc');
		    $buildreq->add_buildreq('perl-pod');
		}
		#$sec->subst_body_if(qr'\s-F(?=\d)',' -F ',qr'^\s*\%patch(?:\d|\s)');
		$sec->map_body(sub {
		    # rpm404 compat: ошибка: строка 37: Неверный аргумент для %patch: %build
		    s,(?<=\s)-b(\S+),-b $1,g if /^\s*\%patch/;
			       });
	    } elsif ($t eq 'build') {
		$sec->map_body(sub {
		    $buildreq->add_buildreq_for('mvn-rpmbuild in %build','maven') if /(?:^|\s)mvn-rpmbuild(?:\s|$)/;
		    $buildreq->add_buildreq_for('waf in %build','waf') if /(?:^|\s)(?:\.\/)?waf(?:\s|$)/;
		    $buildreq->add_buildreq_for('fontforge scripts hack: python core modules are required (#25979)','python') if /fontforge\s.+-script\s/;
		    # redefinition of configure;
		    s,^%global _configure\s,%global _configure_script ,;
			       });
	    } elsif ($t eq 'install') {
		$sec->map_body(sub {
		    # rpm 4.8.0+ has built-in %make_install macro == our %make_install_std
		    s,\%{make_install}(?!\s+install),%{makeinstall_std},;
		    s,\%make_install(\s)(?!\s*install),%makeinstall_std$1,;
		    s,/usr/sbin/hardlink,/usr/bin/hardlink,g;
		    # altlinux firmware poilcy
		    s,\%{_prefix}/lib/firmware/,/lib/firmware/,g;

		    &__convert_fedora_like_macrodir() if $fedora_like_macrodir;
			       } );
	    }
	} elsif ($t eq 'clean') {
	    $sec->exclude($cleanroot_pattern);
	    $sec->delete();
	} elsif ($t eq 'check') {
	    # %check now supported in alt rpm 5.1+
	} elsif ($t eq 'changelog') {
	    if ($opt_import_info) {
		my ($chfh, $chfilename) = tempfile();
		print $chfh 'EVR: ',&RPM::Source::Editor::__evr2string($spec->_get_evr()),"\n";
		for my $tag (qw/Name Version Release Serial Epoch/) {
		    my $value = $mainsec->get_tag($tag);
		    print $chfh $tag, ': ',$value,"\n" if defined $value;
		}
		if ($spec->{'SOURCERPM'}) {
		    print $chfh 'srpm: ',$spec->{'SOURCERPM'},"\n";
		} else {
		    print $chfh 'spec: ',$spec->{'SPECFILE'},"\n";
		}
		print $chfh @{$sec->get_bodyref()};
		close ($chfh);
		$spec->add_source($chfilename, NAME=>'import.info', FILE=>$chfilename);
		unlink $chfilename;
	    }
	    $sec->empty_body();
	} elsif ($t eq 'pre') {
	    &alt_generic_fix_scripts($buildreq,$spec,$sec);
	} elsif ($t eq 'post') {
	    &alt_generic_fix_scripts($buildreq,$spec,$sec);
	} elsif (($t eq 'postun') or ($t eq 'preun')) {
	    &alt_generic_fix_scripts($buildreq,$spec,$sec);
	} elsif ($t eq 'pretrans') {
	    # what to do ???
	} elsif ($t eq 'posttrans') {
	    &alt_generic_fix_scripts($buildreq,$spec,$sec);
	# other scripts == triggers
	} elsif ($RPM::Source::SpecSection::SCRIPT_SECTION{$t}) {
	    # what to do ???
	}
    }

    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

sub __convert_fedora_like_macrodir {
    s,\%{(?:rpm_macros_dir|macrosdir)}/macros\.,%{_rpmmacrosdir}/,g;
    s,\%{(?:rpm_macros_dir|macrosdir)},%{_rpmmacrosdir},g;
};

sub alt_generic_cleanup_headers {
    my ($resourceobj,$spec,$section)=@_;
    my @out;
    my $is_prev_line_multiline=0;
    foreach (@{$section->get_bodyref()}) {

	# in fdupes macro definition
	s!/usr/sbin/hardlink!/usr/bin/hardlink!g;

	if (/^(url|source\d*):(\s*)(.+\S)\s*$/i) {
	    my $Tag=$1;
	    my $tspace=$2;
	    my $url = $3;
	    # TODO: beware of macroses, like Source2: blabla-%{ename}-%bversion.zip

	    # unquoted space in url hack (common in fedora)
	    $url=~s/(?<!\%)\%(2[0-9])/\%\%$1/g;
	    if ($Tag=~/^url$/i) {
		# unquoted unicode in project url hack
		$url=~s/(?<!\%)\%(?=[ABCDEF][0-9])/\%\%/g;
		$url=~s/(?<!\%)\%(?=[0-9][0-9A-F])/\%\%/g;
	    }
	    # beware of shell calls in macros, like
	    # %(tr "." "_" <<<%{version})
	    if ($url !~/\%\(/) {
		# spaces in url, not related to shell calls in macros
		$url=~s/\s(?=\S)/%%20/g;
	    }
	    push @out, "$Tag:$tspace$url\n";
	} elsif (/^orderwithrequires:\s/i) {
	    print STDERR "INFO: mapped to Requires: $_";
	    s/^orderwithrequires:/Requires:/i;
	} elsif (/^release:.*\%mkrel/i) {
	    s/\%mkrel\s+//;
	    $spec->macros()->{release}=~s/\%mkrel\s+// if $spec->macros()->{release};
	    push @out, $_;
	} elsif (/^summary:/i) {
	    push @out, &__kill_non_ascii($_);
	} elsif (/^\s*#/i) {
	    #in comment: conakry-fonts-20070829-4.fc15/*.spec does not look like a spec file
	    push @out, &__kill_non_ascii($_);
	} elsif (/\\\n$/ or $is_prev_line_multiline) {
	    # multiline macro for description: fonts-otf-beteckna
	    push @out, &__kill_non_ascii($_);
	} elsif (/^\s*\%define\s+(?:_binary_payload|_source_payload|_binary_filedigest_algorithm|_source_filedigest_algorithm)\s/i) {
	    # drop! not compatible:
	    # missing mandatory requires: rpmlib\(PayloadIsLzma\).*
	    # girar-task add: mojo-aspectj-maven-plugin-1.4-alt1_1jpp7.src.rpm: sisyphus_check failed
	} else {
	    push @out, $_;
	}
	$is_prev_line_multiline= /\\\n$/;
    }
    $section->set_body(\@out);
}

# is macro %name used in spec before Name:,Version: tags
# it is not allowed in alt
sub alt_generic_check_ahead {
    my ($spec,$section)=@_;
    my @out;
    my ($is_name_passed,$is_name_ahead);
    my ($is_version_passed,$is_version_ahead);
    my ($is_release_passed,$is_release_ahead);
    foreach (@{$section->get_bodyref()}) {
	m/^\s*name:/i and $is_name_passed=1; #and print STDERR "passed $_\n";
	m/^\s*version:/i and $is_version_passed=1; #and print STDERR "passed $_\n";
	m/^\s*release:/i and $is_release_passed=1; #and print STDERR "passed $_\n";
	$is_name_ahead=1 if not $is_name_passed and (m/\%\{name\}/ or m/\%name(?:[^\w\d_]|$)/);
	$is_version_ahead=1 if not $is_version_passed and (m/\%\{version\}/ or m/\%version(?:[^\w\d_]|$)/);
	$is_release_ahead=1 if not $is_release_passed and (m/\%\{release\}/ or m/\%release(?:[^\w\d_]|$)/);
#print STDERR "AHEAD $_\n";
	last if $is_name_passed and $is_version_passed and $is_release_passed;
    }
    if ($is_name_ahead or $is_version_ahead) {
	$section->unshift_body(
	    "# \%\%name or \%\%version is ahead of its definition. Predefining for rpm 4.0 compatibility.\n".
	    '%define name '.$spec->get_section('package','')->get_tag('Name')."\n".
	    '%define version '.$spec->get_section('package','')->get_tag('Version')."\n"
	    );
	print STDERR "INFO: \%name or \%version is ahead of its definition - fixed.\n";
    }
    if ($is_release_ahead) {
	$section->unshift_body(
	    "# \%\%release is ahead of its definition. Predefining for rpm 4.0 compatibility.\n".
	    '%define release '.$spec->get_section('package','')->get_tag('Release')."\n"
	    );
	print STDERR "INFO: \%release is ahead of its definition - fixed.\n";
    }
}

sub alt_generic_fix_desc {
    my ($spec, $section)=@_;
    my @out;
    my $macros=$spec->macros;
#    print STDERR "Generic2ALT: alt_generic_fix_desc: entering.\n" if $verbose>5;
    $section->map_body ( 
	sub {
	    # like, %filter_ and so on...
	    unless (/\%(?:if|ifarch|else|endif|global|define|\{?summary)/ or /^\s*\%/) { # never_quote_macro_commands
		# % in description
		s~(?<!\%)\%(\s|$|[^\%]\S+)~&__quote_macro($macros,$1)~ge;
		$_=&__kill_non_ascii($_);
	    }
	});
#    print STDERR "Generic2ALT: alt_generic_fix_desc: leaving.\n" if $verbose>5;
}

sub __quote_macro {
    my ($macros,$macroexpression)=@_;
    my $raw_macro=$macroexpression;
    $raw_macro=$1 if $macroexpression=~/^\{(.*)\}$/;
    print STDERR "__quote_macro: got $macroexpression\n" if $verbose>3;
    return '%'.$macroexpression if defined $macros->get_raw($raw_macro) or $macros->macro_subst($macroexpression) ne '';
    return '%%'.$macroexpression;
}

sub __kill_non_ascii {
    my ($line)=@_;
    return $line if not $opt_force_ascii;
    local $_=$line;
    # some distro use iso or utf8 in description. ALT Linux use ascii only. :(
    #tr//e/;
    tr/ޜ\x98\x99/-.....eooeaiaoucououaaidieaayvnieouabEOOEAIAOUCOUOUAAIDIEAAYVNIEOUAB''''/;
    # exterminatus of the rest of the non-ascii characters
    s/[\x80-\xff]/./g;
    return $_;
}

sub alt_generic_check_alt_macro {
    my ($spec)=@_;
    my $macros=$spec->macros();
    if ($macros->get_raw('_libdir')) {
	print STDERR "WARNING: %_libdir redefined to ".$macros->get_raw('_libdir')."\n";
    }
    if ($macros->get_raw('homedir') ne $ENV{'HOME'}) {
	my $define_found=0;
	map {$_->map_body(sub {
	    $define_found=1 if s,^(\s*\%(?:global|define)\s+)homedir(\W),${1}apphomedir${2},;
	    s,(?<!\%)(\%{?)homedir(\W|$),${1}apphomedir${2},;
			  }) } $spec->get_sections();
	print STDERR "info: %homedir defined. renamed to %apphomedir\n" if $define_found;
    }
}

sub alt_generic_check_zip {
    my ($buildreq,$spec)=@_;
    my $macros=$spec->macros();
    foreach my $srcitem (@{$spec->raw_sourcelist_ref}) {
	next unless $srcitem;
	$srcitem=$macros->macro_subst($srcitem);
	if ($srcitem=~/\.(?:zip|oxt|xpi)$/i) {
	    $buildreq->add_buildreq_for('one of the sources is a zip file','unzip');
	    return;
	}
    }
    foreach my $secname (qw!prep build install check!) {
	my $sec=$spec->get_section($secname);
	next unless $sec;
	if ($sec->match_body(qr'unzip\s')) {
	    $buildreq->add_buildreq_for(q!required for %!.$secname,'unzip');
	    return;
	}
    }
}

sub alt_generic_check_jni {
    my ($buildreq,$spec)=@_;
    
    if (grep {/\%\{?_(?:jni|java|mavenpom|mavendepmapfrag)dir\}?/} map {@{$_->get_bodyref()}} $spec->get_sections()) {
	$buildreq->add_pre_buildreq_for(q!# required for java specific macros!,'rpm-build-java');
    }
}

sub alt_generic_cleanup_files {
    my ($spec,$section)=@_;
    my %docdir;
    my @out;

    my @f_flags=$section->get_flag('-f');
    if (@f_flags>1) {
	my $base_filelist=shift @f_flags;
	my $installsec=$spec->get_section('install');
	unless ($installsec) {
	    print STDERR "WARN: extra -f flags but no \%install section :-(\n";
	} else {
	    my $bodyref=$section->get_bodyref;
	    foreach (@f_flags) {
		print STDERR "INFO: extra -f flags: merging file list $_ into $base_filelist\n";
		$installsec->push_body("# multiple -f flags in \%files: merging -f $_ into -f $base_filelist
cat $_ >> $base_filelist\n");
		$bodyref->[$section->_get_header_idx]=~s/-f\s+\Q$_\E//;
	    }
	}
    }

    $section->map_body(sub {
	# ALTLinux xmkmf does not create .1x man pages.
	s,\.1x,.1, if m!^\%\{?_mandir\}?/man1/!;
	# due to feature of alt findreq
	s,%{_sysconfdir}/init.d,%_initdir,g;
		       });

    foreach (@{$section->get_bodyref()}) {
	# %caps(cap_ipc_lock=ep)
	s/\%caps\([^\)]+\)//g; # not supported

	s/^\s*\%attr\s*\(-,\s*root,\s*root\s*\)\s*//;
	#if (/^\s*\%defattr\(0644,root,root,0755\)\s*$/
	if (/^\s*\%defattr\s*\((?:0?644|-),\s*root,\s*root,\s*(?:0?755|-)\)\s*$/
	    or /^\s*\%defattr\s*\((?:0?664|-),\s*root,\s*root,\s*(?:0?755|-)\)\s*$/
	    or /^\s*\%defattr\s*\((?:0?644|-),\s*root,\s*root\)\s*$/
	    ) {
	    # throw away
	} elsif (/^\s*\%defattr\s*\(0755,\s*root,\s*root,\s*0755\)\s*$/) {
	    # dangerous: as in antlr may cause lost 755 rights in /usr/bin
	    push @out, "#".$_;
	} elsif (/^\s*\%attr\s*\([24](?:755|751|551),\s*root,/) {
	    # sisyphus_check: unsafe suid/sgid file permissions (readable or writable)
	    s/\%attr\s*\(2755,\s*root,/\%attr\(2711,root,/;
	    s/\%attr\s*\(2751,\s*root,/\%attr\(2711,root,/;
	    s/\%attr\s*\(2551,\s*root,/\%attr\(2711,root,/;
	    s/\%attr\s*\(4755,\s*root,/\%attr\(4711,root,/;
	    s/\%attr\s*\(4751,\s*root,/\%attr\(4711,root,/;
	    s/\%attr\s*\(4551,\s*root,/\%attr\(4711,root,/;
	    push @out, $_;
	} elsif (/^\s*\%doc\s*\%\{_docdir\}\s*$/) {
	    push @out, '%doc %{_docdir}/*'."\n";
	} elsif (m!^\s*\%doc\s*(\%\{?_docdir\}?/\%\{?name\}?-\%\{?version\}?/\S+)!) {
	    &__add_docdir(\%docdir,$1);
	    push @out, $_;
	} elsif (m!^\s*(?:\%doc\s*)?(\%\{?_datadir\}?/doc/\%\{?name\}?-\%\{?version\}?/\S+)!) {
	    &__add_docdir(\%docdir,$1);
	    push @out, $_;
	} elsif (m!^\s*(\%\{?_docdir\}?/\%\{?name\}?-\%\{?version\}?/\S+)!) {
	    &__add_docdir(\%docdir,$1);
	    # note: no need to add a %doc here, rpm will add it for us.
	    push @out, $_;
	} elsif (/^\s*\%doc\s+.*"/) {
	    s/"([^"]+)"/&__doc_quote_space($1)/ge;
	    push @out, $_;
	} elsif (m!^\s*((?:\%lang\(\S+\)\s*)?\%\{?_mandir\}?/\S+)\.[gx]z\s*$!) {
	    # man pages can be auto-bzipped or auto-uncompressed
	    push @out, $1."*\n";
	} elsif (m!^\s*((?:\%lang\(\S+\)\s*)?(?:/usr/share|\%\{?_datadir\}?)/man/man\d/\S+)\.[gx]z\s*$!) {
	    # man pages can be auto-bzipped or auto-uncompressed
	    push @out, $1."*\n";
	} elsif (m!^\s*(\%\{?_infodir\}?/\S+)\.[gx]z\s*$!) {
	    # info can be auto-bzipped
	    push @out, $1.".*\n";
	} elsif (m!^\s*\%{?_datadir}?/icons/\*\s*$!) {
	    #sisyphus_check: check-intersects ERROR: intersections with system packages /usr/share/icons/hicolor/48x48/apps
	    push @out, "\%_iconsdir/*/*/*/*\n";
	} elsif (m!^\s*\%{?_datadir}?/icons/hicolor(?:/\*)?\s*$!) {
	    #sisyphus_check: check-intersects ERROR: intersections with system packages /usr/share/icons/hicolor/48x48/apps
	    push @out, "\%_iconsdir/hicolor/*/*/*\n";
	} elsif (m!^\s*\%{?_datadir}?/locale/(?:\*)?\s*$!) {
	    #sisyphus_check: check-intersects ERROR: intersections with glibc locale
	    push @out, "\%_datadir/locale/*/*\n";
	} elsif (m!^\s*(?:\%exclude)?\s*\%{?_libdir}?/[^/]+\.la\s*$!) {
	    # skip the la in _libdir
	} elsif (m!^(\s*\%{?_libdir}?/pkgconfig)\s*$!) {
	    # ownership violation in _libdir
	    push @out, "$1/*\n";
	} elsif (m!^\s*\%{?_libdir}?/qt4/mkspecs!) {
	    s!\%{?_libdir}?/qt4/mkspecs!%{_datadir}/qt4/mkspecs!;
	    push @out, $_;
	} elsif (m!^\s*\%{?_libdir}?/\*\s*$!) {
	    push @out, $_;
	    push @out, '# wildcard _libdir/*'."\n".
		'%exclude %_prefix/lib/debug'."\n";
	} elsif (m!^\s*\%{?_libdir}?/lib\*la\s*$!) {
	    # skip; it will be autocleaned
	} elsif (m!^\s*\%{?_libexecdir}?/\*\s*$!) {
	    push @out, $_;
	    # _libexecdir redefined ?
	    unless ($spec->macros()->get_raw('_libexecdir')) {
		push @out, '# wildcard _libexecdir/*'."\n".
		'%exclude %_prefix/lib/debug'."\n";
	    }
	} elsif (m!^\s*\%{?_mandir}?/?\s*$!) {
	    push @out, "%{_mandir}/*/*\n";
	} elsif (m!^\s*\%{_prefix}/lib/firmware/!) {
	    s,\%{_prefix}/lib/firmware/,/lib/firmware/,;
	    push @out, $_;
	} else {
	    push @out, $_;
	}
    }
    $section->set_body(\@out);
    if (scalar %docdir) {
	$section->push_body('# hack; explicitly added docdir if not owned
'.join ('',map {'%doc %dir '.$_."\n"} keys (%docdir)) );
    }
}

sub __add_docdir {
    my ($dochashptr,$dir) =@_;
    $dir=~s/\%name-\%version/%{name}-%{version}/;
    $dir=~s/\%_docdir/%{_docdir}/;
    $dir=~s!\%{_datadir}/doc!%{_docdir}!;
    $dir=~s!\%_datadir/doc!%{_docdir}!;
    $dochashptr->{dirname($dir)}=1;
    $dochashptr->{'%{_docdir}/%{name}-%{version}'}=1;
}

sub __doc_quote_space {
    my $word=$1;
    $word=~s/(\s)/\\$1/g;
    return $word;
}

# --------------------------------------------------------------
# script conversion code;
# --------------------------------------------------------------
# TODO: have a heart to refactor it someday

our $debug_fix_scripts=0;

sub alt_generic_fix_scripts {
    my ($buildreq,$spec,$section)=@_;
    my $type=$section->get_type;
    my $pkg=$section->get_canonical_package;
    print STDERR "entering section $type $pkg:\n" if $debug_fix_scripts>1;
    my $shell=$section->get_flag('-p');
    if (defined $shell) {
	if ($shell=~/(ldconfig|gtk-update-icon-cache)/) {
	    print STDERR 'Generic2ALT: deleted %'.$type." ".$pkg." -p *${1}*\n" if $RPM::Source::Transform::verbose>1;
	    $section->delete();
	    return;
	} elsif ($shell=~/<lua>/) {
	    $section->map_body(sub {
		s!posix.utime\("\%{_datadir}/icons/hicolor"\)!!;
			       });

	    if ($section->is_empty()) {
		$section->delete();
	    } else {
		print STDERR 'Generic2ALT: %'.$section->get_type()." ".$section->get_raw_package()." contains unrecognized lua code. Please, convert manually :( \n";
	    }
	}
    }
    print STDERR "before_main_map_body:\n",@{$section->get_bodyref()} if $debug_fix_scripts>1;
    # another multliline hack
    my $multiline_flag=0;
    my $drop_flag=0;
    $section->map_body(sub {
	if (/^\s*setcap\s/) {
	    chomp;
	    $_='[ -x /sbin/setcap ] && '.$_." ||:\n";
	    return;
	}
	$multiline_flag= /\\\n$/s;
	if ($drop_flag) {
	    $_='';
	    $drop_flag=$multiline_flag;
	    return;
	}

	# %gconf_schema_* family conversion
	#%pre                                                                            
	#%gconf_schema_prepare %{name}                                                   
	#%post                                                                           
	#%gconf_schema_upgrade %{name}                                                   
	#%preun                                                                          
	#%gconf_schema_remove %{name} 
	s,^\s*\%gconf_schema_prepare\s+\S*,,;
	s,\%gconf_schema_upgrade(\s+\S*),\%gconf2_install$1,;
	if (s,^\s*\%gconf_schema_remove(\s+\S*),,) {
	    $_='if [ "$1" -eq 0 ]; then'."\n".
		"    \%gconf2_uninstall $1\n".
		"fi\n";
	    $buildreq->add_pre_buildreq('libGConf-devel');
	    return;
	}

	####################################################
	# dot replacers
	####################################################
	s,/sbin/ldconfig,: ,;

	# deprecated macros
	# make_session is update_wms from mageia
	# update/clean_icon_cache is from mageia
	s,^\s*\%(?:update_wms|make_session|update_icon_cache|clean_icon_cache|update_maven_depmap)(\s|\$),: ,g;
	# should be performed together - order is important! systemd_post matches both re
	if (s,^\s*\%systemd_(post|preun)\s+,,) {
	    my $prefix=$1;
	    my $out='';
	    while (s,^\s*(\S+)\.service,,) {
		$out.='%'.$prefix.'_service '.$1."\n";
	    }
	    $_=$out;
	    return;
	}
	s,^\s*:\s+\S+\.service,: ,g if s,^\s*\%systemd_post(?:un)?(?:_with_restart)?\s+\S+\.service,: ,g;

	# tex policy
	s,(?:/usr/bin/|/sbin/|/usr/sbin/|\%\{?_s?bindir\}?/?)?texhash([^;]+|$),: ,;

	# those below are obsolete by alt filetriggers, like
	# update-desktop-database|install-info|scrollkeeper-update (obsolete in alt)
	$drop_flag=$multiline_flag if
	    s,^\s*(?:\[\s+-[xef]\s+)?(?:/usr/bin/|/bin/|/sbin/|/usr/sbin/|\%\{?_s?bindir\}?/?)?(update-desktop-database|install-info|scrollkeeper-update|glib-compile-schemas|update-mime-database|systemctl|update-gtk-immodules|gtk-update-icon-cache|gtk-query-immodules-\S+)(\s[^;]*|\$),,;


	####################################################
	# excluders
	####################################################
	if (
	    # obsolete-call-in-post-gtk-update-icon-cache.pl (repocop-unittest-rpm-filetriggers-0.20-alt7)
	    m!^\s*(?:\%\{?_bindir\}?/|/bin/|/usr/bin/)?touch\s+--no-create\s+((?:\%\{?_datadir\}?|/usr/share)/icons|\%\{?_(?:kde4_|K4)?iconsdir\}?)/[\w\d_]+! or
	    # deprecated by gio filetrigger
	    m!^\s*(?:\%{?_bindir}?/)?gio-querymodules(?:-\%{__isa_bits})?\s+\%{?_libdir}?/gio/modules!
	    ) {
	    $_='';
	    return;
	}

	####################################################
	# modificators
	####################################################
	# TODO: discuss whether to implement or drop.
	# right now, w/o ||: it will fail during install
	s,(/sbin/udevadm\s+control\s+--reload-rules)(\s*)$,$1 ||:$2,;

		       });
    print STDERR "after_main_map_body:\n",@{$section->get_bodyref()} if $debug_fix_scripts>1;

    # tex policy
    $section->multi_exclude_body(
	qr'^\s*conffile="\$\(texconfig',
	qr'^\s*if\s*\[\s+"?\$1"?\s+(?:[-]eq|=)\s+"?\d+"?\s+]\s*;\s+then\s*(?:#|$)',
	qr'^\s*.*updmap',
	qr'^\s*fi\s*$',
	);

    # TODO: service chkconfig (after repocop fix)?

    $section->multi_exclude_body(
	qr'^\s*if\s+\[\s+-x\s+.*fc-cache\s+\]\s*;\s+then',
	qr'^\s*.*fc-cache',
	qr'^\s*fi',
	);

    #eclipse-reconciler.sh %{_libdir}/eclipse %{_tmppath} > /dev/null
    unless ($type eq 'post' and $pkg eq 'platform') { # eclipse-platform
	$section->exclude(qr'^\s*eclipse-reconciler.sh\s+\%{?_libdir}?/(?:\%{name}|eclipse)');
	print STDERR 'in %'."$type $pkg: should kill eclipse-reconciler.sh\n" if $section->match_body('eclipse-reconciler.sh');
    }

    # TODO: merge with code below
    $section->multi_exclude_body(
	qr'^\s*if\s+\[\s+-x\s+.*gtk-update-icon-cache\s+\]\s*;\s*then',
	qr'^\s*.*gtk-update-icon-cache',
	qr'^\s*fi',
	);
    $section->multi_exclude_body(
	qr'^\s*\[\s+"?\$1"?\s+-eq\s+1\s+\] &&\s+\\$',
	qr'^\s*.*gtk-update-icon-cache',
	);

    #---------------------------------------------------------------------


    # below is just a bunch of hacks :(
    &RPM::Source::Convert::Plugins::Generic2ALT::_generic_filter_script_bin_true($section);
    $section->multi_exclude_body(
qr'^\s*if\s+\[\s+-x\s+\%{_bindir}/rebuild-gcj-db\s+\]\s*$',
qr'^\s*then\s*(?:#|$)',
#%{_bindir}/rebuild-gcj-db
qr'/rebuild-gcj-db\s*$',
qr'^\s*fi\s*$');
    $section->exclude(qr'^\s*if\s+\[\s+-x\s+\%\{?_bindir\}?/rebuild-gcj-db\s+\];\s*then\s+\%\{?_bindir\}?/rebuild-gcj-db\s*\|\|\s*true\s*;\s*fi\s*$');
#{
#  [ $1 -eq 0 ] || exit 0
#OR  [ $1 = 0 ] || exit 0
#  :
#} >/dev/null 2>&1 || :
    $section->multi_exclude_body(
qr'^\s*\{\s*$',
qr'^\s*\[\s+"?\$1"?\s+(?:-eq|-ge|-ne|-gt|=)\s+"?[01]"?\s+\]\s*\|\|\s*exit\s+0\s*$',
qr'^\s*(?:#.*|:\s*|\s*)$',
qr'^\s*\}\s*\>\s*/dev/null\s+2\>\&1\s*\|\|\s*:\s*$');

    # cleaning up the mess
    while (&_clean_up_empty_tests($section)) {};
    while (&_clean_up_ifs_iteration($section)) {};
    while (&_clean_up_for_iteration($section)) {};
    while (&_clean_up_ifs_iteration($section)) {};

    # TODO: merge
    {
	# hack: filter to get rid of : as placeholders
	my @body=@{$section->get_bodyref()};
	my @new_body;
	# header
	push @new_body, shift @body;
	my $no_meaningful_lines=1;
	while (scalar @body) {
	    local $_=shift(@body);
	    next if /^\s*:[^;]*$/ && $no_meaningful_lines;
	    next if /^\s*#/ && $no_meaningful_lines;
	    next if /^\s*$/ && $no_meaningful_lines;
	    # [ $1 -eq 0 ] || exit 0
	    next if /^\s*\[\s+"?\$1"?\s+(?:-eq|-ge|-ne|-gt|=)\s+"?\d+"?\s+]\s*(?:\&\&|\|\|)\s*exit\s+0\s*$/ && $no_meaningful_lines;
	    # exit 0
	    next if /^\s*exit\s+0\s*$/ && $no_meaningful_lines;
	    $no_meaningful_lines=0 unless /^\s*$/ or /^\s*#/;
	    push @new_body, $_;
	}
	$section->set_body(\@new_body);
    }


    # hack again!!!
    $section->set_body(&RPM::Source::SpecSection::__filter_empty_macro_if_endif($section->get_bodyref()));
    &RPM::Source::Convert::Plugins::Generic2ALT::_generic_filter_script_bin_true($section);
    print STDERR "Oops! Generic2ALT: internal: dead 6: %",$section->get_type()," ",$section->get_raw_package(),"\n" unless (@{$section->get_bodyref()});

    if ($section->is_empty()) {
	print STDERR 'Generic2ALT: deleted empty %'.$section->get_type()." ".$section->get_raw_package()."\n" if ($RPM::Source::Transform::verbose>1);
	$section->delete();
    } else {
	print STDERR 'info: Generic2ALT: keep %'.$section->get_type()." ".$section->get_raw_package()."\n";
    }
}

sub _clean_up_empty_tests {
    my ($section)=@_;
    my $succeeded=0;
    my @body=@{$section->get_bodyref()};
    my (@new_body,@new_body_test);
    # header
    push @new_body, shift @body;
    my $in_test=0;
    my $multiline=0;
    my $meaningful_lines=0;
    while (scalar @body) {
	local $_=shift(@body);
	if ($multiline && $in_test) {
	    $multiline=/\\$/;
	    push @new_body_test, $_;
	    next if /^\s*#/;
	    next if /^\s*$/;
	    next if /^\s*:[^;]*$/;
	    next if /^\s*exit\s+0\s*$/;
	    $meaningful_lines=1;
	} elsif (/^\s*\[\s*-[exf]\s+\S+\s*\]\s*(?:\&\&|\|\|)\s*\\$/
	    or /^\s*\[\s+"?\$1"?\s+(?:-eq|-ge|-ne|-gt|=)\s+"?\d+"?\s+]\s*(?:\&\&|\|\|)\s*\\$/
	    or /^\s*\[\s*-[exf]\s+\S+\s*\]\s*(?:\&\&|\|\|)\s*\[\s+"?\$1"?\s+(?:-eq|-ge|-ne|-gt|=)\s+"?\d+"?\s+]\s*(?:\&\&|\|\|)\s*\\$/
	    ) {
	    push @new_body_test, $_;
	    $meaningful_lines=0;
	    $multiline=1;
	    $in_test=1;
	} else {
	    $multiline=0;
	    $in_test=0;
	    $meaningful_lines=0;
	    push @new_body, @new_body_test if $meaningful_lines;
	    push @new_body, $_;
	    @new_body_test=();
	}
    }
    $section->set_body(\@new_body);
    return 0;
}

sub _clean_up_ifs_iteration {
    my ($section)=@_;
    my $succeeded=0;
    #if [ $1 -eq 0 ]; then
    #if [ "$1" = "0" ]; then
    #fi
    # hack: filter to get rid of : as placeholders
    my @body=@{$section->get_bodyref()};
    my (@new_body_if,@new_body_tail);
    # header /@body>1/
    while (scalar @body>1) {
	@new_body_if=();
	while (scalar @body>1) {
	    local $_=pop(@body);
	    unless (
		/^\s*if\s*\[\s*"?\$\{?\d+\}?"?\s+(?:[-](?:eq|gt|ge|ne)|=)\s+"?\d+"?\s+\]\s*(?:;\s+then\s*)?(?:#|$)/ or
		m!^\s*if\s*\(\s*update-mime-database\s+-v\s+\&\>\s*/dev/null\s*\)(?:;\s+then\s*)?(?:#|$)! or
		# empty test
		m!^\s*if\s*\[\s+\-\w\s+\S+\s+\]\s*(?:;\s+then\s*)?(?:#|$)!
		) {
		unshift @new_body_tail, $_;
	    } else {
		push @new_body_if, $_;
		print STDERR "fixscripts: started new if: $_" if $debug_fix_scripts;
		last;
	    }
	}
	if (@new_body_if>0) {
	    my $meaningful_lines=0;
	    while (scalar @new_body_tail) {
		local $_=shift(@new_body_tail);
		push @new_body_if, $_;
		last if /^\s*fi(?:\s+\|\|\s*:)?\s*$/;
		unless (/^\s*:[^;]*$/ or /^\s*then\s*(?:#|$)/ or /^\s*#/ or /^\s*$/ or /^\s*exit\s+0\s*$/) {
		    $meaningful_lines=1;
		    print STDERR "fixscripts: if: found meaningful_lines: $_" if $debug_fix_scripts>2;
		}
	    }
	    print STDERR "fixscripts: if: meaningful_lines=$meaningful_lines\n\tnew_body_if has ".@new_body_if."lines:\n",@new_body_if,"\tnew_body_tail has ".@new_body_tail."lines:\n",@new_body_tail if $debug_fix_scripts>1;
	    unless ($meaningful_lines) {
		@new_body_if=();
		$succeeded=1;
	    }
	}
	unshift @new_body_tail, @new_body_if;
    }
    $section->set_body([@body,@new_body_tail]);
    return $succeeded;
}

#  for s in %services; do                                                        
#    # Package upgrade, not uninstall                                            
#  done                                                                          
sub _clean_up_for_iteration {
    my ($section)=@_;
    my $succeeded=0;
    # hack: filter to get rid of : as placeholders
    my @body=@{$section->get_bodyref()};
    my (@new_body_for,@new_body_tail);
    # header /@body>1/
    while (scalar @body>1) {
	@new_body_for=();
	while (scalar @body>1) {
	    local $_=pop(@body);
	    unless (
		/^\s*for\s*\S+\s+in\s+[^;#]+(?:;\s+do\s*)?(?:#|$)/
		) {
		unshift @new_body_tail, $_;
	    } else {
		push @new_body_for, $_;
		print STDERR "fixscripts: started new for: $_" if $debug_fix_scripts;
		last;
	    }
	}
	if (@new_body_for>0) {
	    my $meaningful_lines=0;
	    while (scalar @new_body_tail) {
		local $_=shift(@new_body_tail);
		push @new_body_for, $_;
		if (/^\s*done\s*$/) {
		    print STDERR "fixscripts: for - done: meaningful_lines=$meaningful_lines\n\tnew_body_for has ".@new_body_for."lines:\n",@new_body_for if $debug_fix_scripts>1;
		    last;
		}
		unless (/^\s*:[^;]*$/ or /^\s*do\s*(?:#|$)/ or /^\s*#/ or /^\s*$/ or /^\s*exit\s+0\s*$/) {
		    $meaningful_lines=1;
		}
	    }
	    print STDERR "fixscripts: for: meaningful_lines=$meaningful_lines\n\tnew_body_for has ".@new_body_for."lines:\n",@new_body_for,"\tnew_body_tail has ".@new_body_tail."lines:\n",@new_body_tail if $debug_fix_scripts>1;
	    unless ($meaningful_lines) {
		@new_body_for=();
		$succeeded=1;
	    }
	}
	unshift @new_body_tail, @new_body_for;
    }
    $section->set_body([@body,@new_body_tail]);
    return $succeeded;
}


sub _generic_filter_script_bin_true {
    my ($section)=@_;
    my @header;
    my @out;
    my $i = 0;
    my $command_found=0;
    unless (@{$section->get_bodyref()}) {
	print STDERR "strange dead section: %",$section->get_type()," ",$section->get_raw_package(),"\n";
	return;
    }
    push @header, shift @{$section->get_bodyref()};
    foreach (@{$section->get_bodyref()}) {
	if (/^\s*:\s*$/) {
	    push @out, $_ if $i++==0 and $command_found;
	}elsif (/^\s*#/) {
	    push @out, $_;
	}elsif (/^\s*$/) {
	    push @out, $_;
	} else {
	    push @out, $_;
	    #print STDERR "other for ".$section->get_type()," ",$section->get_raw_package(),"\n",$_;
	    $command_found=1;
	    $i=0;
	}
    }
    if ($command_found) {
	$section->set_body([@header,@out]);
	#print STDERR "set cmd for ",@header;
    } else {
	$section->set_body([@header]);
	#print STDERR "Jpackage2ALT: set empty for ",@header;
    }
}


package Generic2ALT::DependencyFilter::PostScriptsDepsCleanup;

use strict;
use warnings;
our @ISA=q/RPM::Source::Tools::Dependency::Filter/;

sub is_applicable {
    my ($self,$tagobj)=@_;
    return 0 unless $self->SUPER::is_applicable($tagobj);
    return $tagobj->flagbits & (
	  RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_PRE
	| RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_POST
	| RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_PREUN
	| RPM::Source::Tools::Dependency::Filter::SPEC_FLAG_SCRIPT_POSTUN
	);
}


sub apply {
    my ($self, $tagobj, $dep) = @_;
    my $pkgname=$dep->{-name};
    return if grep {$pkgname eq $_} qw!systemd chkconfig alternatives %{_sbindir}/alternatives
/sbin/install-info
!;
    return $dep;
}

1;
